@@ -54,6 +54,21 @@
 
 /*--------------------------------------------------------------------*/
 
+enum connwait_e {
+	CW_DO_CONNECT = 1,
+	CW_QUEUED,
+	CW_DEQUEUED,
+	CW_BE_BUSY,
+};
+
+struct connwait {
+	unsigned			magic;
+#define CONNWAIT_MAGIC			0x75c7a52b
+	enum connwait_e			cw_state;
+	VTAILQ_ENTRY(connwait)		cw_list;
+	pthread_cond_t			cw_cond;
+};
+
 static const char * const vbe_proto_ident = "HTTP Backend";
 
 static struct lock backends_mtx;
@@ -99,12 +114,98 @@ VBE_Connect_Error(struct VSC_vbe *vsc, int err)
 	do {								\
 		CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);			\
 		dst = bo->tmx;						\
-		if (dst == 0.0)						\
+		if (isnan(dst) && be->tmx >= 0.0)			\
 			dst = be->tmx;					\
-		if (dst == 0.0)						\
+		if (isnan(dst))						\
 			dst = cache_param->tmx;				\
 	} while (0)
 
+#define FIND_BE_SPEC(tmx, dst, be, def)					\
+	do {								\
+		CHECK_OBJ_NOTNULL(bp, BACKEND_MAGIC);			\
+		dst = be->tmx;						\
+		if (dst == def)						\
+			dst = cache_param->tmx;				\
+	} while (0)
+
+#define FIND_BE_PARAM(tmx, dst, be)					\
+	FIND_BE_SPEC(tmx, dst, be, 0)
+
+#define FIND_BE_TMO(tmx, dst, be)					\
+	FIND_BE_SPEC(tmx, dst, be, -1.0)
+
+#define BE_BUSY(be)	\
+	(be->max_connections > 0 && be->n_conn >= be->max_connections)
+
+/*--------------------------------------------------------------------*/
+
+int
+VBE_is_ah_auto (const struct backend *bp)
+{
+	CHECK_OBJ_NOTNULL(bp, BACKEND_MAGIC);
+	return (bp->director->vdir->admin_health == VDI_AH_AUTO);
+}
+
+void
+VBE_connwait_signal_all(const struct backend *bp)
+{
+	struct connwait *cw;
+	unsigned wait_limit;
+	vtim_dur wait_tmod;
+
+	CHECK_OBJ_NOTNULL(bp, BACKEND_MAGIC);
+
+	FIND_BE_PARAM(backend_wait_limit, wait_limit, bp);
+	FIND_BE_TMO(backend_wait_timeout, wait_tmod, bp);
+
+	if (wait_limit == 0 || wait_tmod <= 0)
+		return;
+
+	Lck_Lock(bp->director->mtx);
+	VTAILQ_FOREACH(cw, &bp->cw_head, cw_list) {
+		CHECK_OBJ(cw, CONNWAIT_MAGIC);
+		assert(cw->cw_state == CW_QUEUED);
+		PTOK(pthread_cond_signal(&cw->cw_cond));
+	}
+	Lck_Unlock(bp->director->mtx);
+}
+
+static void
+vbe_connwait_signal_locked(const struct backend *bp)
+{
+	struct connwait *cw;
+
+	Lck_AssertHeld(bp->director->mtx);
+
+	if (bp->n_conn < bp->max_connections) {
+		cw = VTAILQ_FIRST(&bp->cw_head);
+		if (cw != NULL) {
+			CHECK_OBJ(cw, CONNWAIT_MAGIC);
+			assert(cw->cw_state == CW_QUEUED);
+			PTOK(pthread_cond_signal(&cw->cw_cond));
+		}
+	}
+}
+
+static void
+vbe_connwait_dequeue_locked(struct backend *bp, struct connwait *cw)
+{
+	Lck_AssertHeld(bp->director->mtx);
+	assert(cw->cw_state == CW_QUEUED);
+	VTAILQ_REMOVE(&bp->cw_head, cw, cw_list);
+	vbe_connwait_signal_locked(bp);
+	cw->cw_state = CW_DEQUEUED;
+}
+
+static void
+vbe_connwait_fini(struct connwait *cw)
+{
+	CHECK_OBJ_NOTNULL(cw, CONNWAIT_MAGIC);
+	assert(cw->cw_state != CW_QUEUED);
+	PTOK(pthread_cond_destroy(&cw->cw_cond));
+	FINI_OBJ(cw);
+}
+
 /*--------------------------------------------------------------------
  * Get a connection to the backend
  *
@@ -121,6 +222,10 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 	vtim_dur tmod;
 	char abuf1[VTCP_ADDRBUFSIZE], abuf2[VTCP_ADDRBUFSIZE];
 	char pbuf1[VTCP_PORTBUFSIZE], pbuf2[VTCP_PORTBUFSIZE];
+	unsigned wait_limit;
+	vtim_dur wait_tmod;
+	vtim_dur wait_end;
+	struct connwait cw[1];
 
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CHECK_OBJ_NOTNULL(ctx->bo, BUSYOBJ_MAGIC);
@@ -135,20 +240,56 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 		VSC_C_main->backend_unhealthy++;
 		return (NULL);
 	}
+	INIT_OBJ(cw, CONNWAIT_MAGIC);
+	PTOK(pthread_cond_init(&cw->cw_cond, NULL));
+	Lck_Lock(bp->director->mtx);
+	FIND_BE_PARAM(backend_wait_limit, wait_limit, bp);
+	FIND_BE_TMO(backend_wait_timeout, wait_tmod, bp);
+	cw->cw_state = CW_DO_CONNECT;
+	if (!VTAILQ_EMPTY(&bp->cw_head) || BE_BUSY(bp))
+		cw->cw_state = CW_BE_BUSY;
 
-	if (bp->max_connections > 0 && bp->n_conn >= bp->max_connections) {
+	if (cw->cw_state == CW_BE_BUSY && wait_limit > 0 &&
+	    wait_tmod > 0.0 && bp->cw_count < wait_limit) {
+		VTAILQ_INSERT_TAIL(&bp->cw_head, cw, cw_list);
+		bp->cw_count++;
+		VSC_C_main->backend_wait++;
+		cw->cw_state = CW_QUEUED;
+		wait_end = VTIM_real() + wait_tmod;
+		do {
+			err = Lck_CondWaitUntil(&cw->cw_cond, bp->director->mtx,
+			    wait_end);
+		} while (err == EINTR);
+		bp->cw_count--;
+		if ((err != 0 && BE_BUSY(bp)) || !VRT_Healthy(ctx, dir, NULL)) {
+			VTAILQ_REMOVE(&bp->cw_head, cw, cw_list);
+			VSC_C_main->backend_wait_fail++;
+			cw->cw_state = CW_BE_BUSY;
+		}
+	}
+	Lck_Unlock(bp->director->mtx);
+
+	if (cw->cw_state == CW_BE_BUSY) {
 		VSLb(bo->vsl, SLT_FetchError,
 		     "backend %s: busy", VRT_BACKEND_string(dir));
 		bp->vsc->busy++;
 		VSC_C_main->backend_busy++;
+		vbe_connwait_fini(cw);
 		return (NULL);
 	}
 
 	AZ(bo->htc);
 	bo->htc = WS_Alloc(bo->ws, sizeof *bo->htc);
+	/* XXX: we may want to detect the ws overflow sooner */
 	if (bo->htc == NULL) {
 		VSLb(bo->vsl, SLT_FetchError, "out of workspace");
 		/* XXX: counter ? */
+		if (cw->cw_state == CW_QUEUED) {
+			Lck_Lock(bp->director->mtx);
+			vbe_connwait_dequeue_locked(bp, cw);
+			Lck_Unlock(bp->director->mtx);
+		}
+		vbe_connwait_fini(cw);
 		return (NULL);
 	}
 	bo->htc->doclose = SC_NULL;
@@ -165,6 +306,12 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 		     VRT_BACKEND_string(dir), err, VAS_errtxt(err));
 		VSC_C_main->backend_fail++;
 		bo->htc = NULL;
+		if (cw->cw_state == CW_QUEUED) {
+			Lck_Lock(bp->director->mtx);
+			vbe_connwait_dequeue_locked(bp, cw);
+			Lck_Unlock(bp->director->mtx);
+		}
+		vbe_connwait_fini(cw);
 		return (NULL);
 	}
 
@@ -177,6 +324,9 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 	bp->n_conn++;
 	bp->vsc->conn++;
 	bp->vsc->req++;
+	if (cw->cw_state == CW_QUEUED)
+		vbe_connwait_dequeue_locked(bp, cw);
+
 	Lck_Unlock(bp->director->mtx);
 
 	CHECK_OBJ_NOTNULL(bo->htc->doclose, STREAM_CLOSE_MAGIC);
@@ -198,7 +348,9 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 		bp->n_conn--;
 		bp->vsc->conn--;
 		bp->vsc->req--;
+		vbe_connwait_signal_locked(bp);
 		Lck_Unlock(bp->director->mtx);
+		vbe_connwait_fini(cw);
 		return (NULL);
 	}
 	bo->acct.bereq_hdrbytes += err;
@@ -217,6 +369,7 @@ vbe_dir_getfd(VRT_CTX, struct worker *wrk, VCL_BACKEND dir, struct backend *bp,
 	    bo->htc->first_byte_timeout, bo, bp);
 	FIND_TMO(between_bytes_timeout,
 	    bo->htc->between_bytes_timeout, bo, bp);
+	vbe_connwait_fini(cw);
 	return (pfd);
 }
 
@@ -240,7 +393,7 @@ vbe_dir_finish(VRT_CTX, VCL_BACKEND d)
 	bo->htc->priv = NULL;
 	if (bo->htc->doclose != SC_NULL || bp->proxy_header != 0) {
 		VSLb(bo->vsl, SLT_BackendClose, "%d %s close %s", *PFD_Fd(pfd),
-		    VRT_BACKEND_string(d), bo->htc->doclose->desc);
+		    VRT_BACKEND_string(d), bo->htc->doclose->name);
 		VCP_Close(&pfd);
 		AZ(pfd);
 		Lck_Lock(bp->director->mtx);
@@ -258,6 +411,7 @@ vbe_dir_finish(VRT_CTX, VCL_BACKEND d)
 	bp->vsc->conn--;
 #define ACCT(foo)	bp->vsc->foo += bo->acct.foo;
 #include "tbl/acct_fields_bereq.h"
+	vbe_connwait_signal_locked(bp);
 	Lck_Unlock(bp->director->mtx);
 	bo->htc = NULL;
 }
@@ -366,6 +520,7 @@ vbe_dir_http1pipe(VRT_CTX, VCL_BACKEND d)
 	struct backend *bp;
 	struct v1p_acct v1a;
 	struct pfd *pfd;
+	vtim_real deadline;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
@@ -381,21 +536,26 @@ vbe_dir_http1pipe(VRT_CTX, VCL_BACKEND d)
 
 	ctx->req->res_mode = RES_PIPE;
 
+	retval = SC_TX_ERROR;
 	pfd = vbe_dir_getfd(ctx, ctx->req->wrk, d, bp, 0);
 
-	if (pfd == NULL) {
-		retval = SC_TX_ERROR;
-	} else {
+	if (pfd != NULL) {
 		CHECK_OBJ_NOTNULL(ctx->bo->htc, HTTP_CONN_MAGIC);
 		i = V1F_SendReq(ctx->req->wrk, ctx->bo,
 		    &v1a.bereq, &v1a.out);
 		VSLb_ts_req(ctx->req, "Pipe", W_TIM_real(ctx->req->wrk));
-		if (i == 0)
-			V1P_Process(ctx->req, *PFD_Fd(pfd), &v1a);
+		if (i == 0) {
+			deadline = ctx->bo->task_deadline;
+			if (isnan(deadline))
+				deadline = cache_param->pipe_task_deadline;
+			if (deadline > 0.)
+				deadline += ctx->req->sp->t_idle;
+			retval = V1P_Process(ctx->req, *PFD_Fd(pfd), &v1a,
+			    deadline);
+		}
 		VSLb_ts_req(ctx->req, "PipeSess", W_TIM_real(ctx->req->wrk));
-		ctx->bo->htc->doclose = SC_TX_PIPE;
+		ctx->bo->htc->doclose = retval;
 		vbe_dir_finish(ctx, d);
-		retval = SC_TX_PIPE;
 	}
 	V1P_Charge(ctx->req, &v1a, bp->vsc);
 	CHECK_OBJ_NOTNULL(retval, STREAM_CLOSE_MAGIC);
@@ -449,6 +609,7 @@ vbe_free(struct backend *be)
 #undef DN
 	free(be->endpoint);
 
+	assert(VTAILQ_EMPTY(&be->cw_head));
 	FREE_OBJ(be);
 }
 
@@ -468,8 +629,9 @@ vbe_panic(const struct director *d, struct vsb *vsb)
 {
 	struct backend *bp;
 
-	CHECK_OBJ_NOTNULL(d, DIRECTOR_MAGIC);
-	CAST_OBJ_NOTNULL(bp, d->priv, BACKEND_MAGIC);
+	PAN_CheckMagic(vsb, d, DIRECTOR_MAGIC);
+	bp = d->priv;
+	PAN_CheckMagic(vsb, bp, BACKEND_MAGIC);
 
 	VCP_Panic(vsb, bp->conn_pool);
 	VSB_printf(vsb, "hosthdr = %s,\n", bp->hosthdr);
@@ -566,7 +728,7 @@ VRT_backend_vsm_need(VRT_CTX)
  * The new_backend via parameter is a VCL_BACKEND, but we need a (struct
  * backend)
  *
- * For now, we resolve it when creating the backend, which imples no redundancy
+ * For now, we resolve it when creating the backend, which implies no redundancy
  * / load balancing across the via director if it is more than a simple backend.
  */
 
@@ -680,6 +842,7 @@ VRT_new_backend_clustered(VRT_CTX, struct vsmw_cluster *vc,
 	ALLOC_OBJ(be, BACKEND_MAGIC);
 	if (be == NULL)
 		return (NULL);
+	VTAILQ_INIT(&be->cw_head);
 
 #define DA(x)	do { if (vrt->x != NULL) REPLACE((be->x), (vrt->x)); } while (0)
 #define DN(x)	do { be->x = vrt->x; } while (0)
@@ -687,6 +850,18 @@ VRT_new_backend_clustered(VRT_CTX, struct vsmw_cluster *vc,
 #undef DA
 #undef DN
 
+#define CPTMO(a, b, x) do {				\
+		if ((a)->x < 0.0 || isnan((a)->x))	\
+			(a)->x = (b)->x;		\
+	} while(0)
+
+	if (viabe != NULL) {
+		CPTMO(be, viabe, connect_timeout);
+		CPTMO(be, viabe, first_byte_timeout);
+		CPTMO(be, viabe, between_bytes_timeout);
+	}
+#undef CPTMO
+
 	if (viabe || be->hosthdr == NULL) {
 		if (vrt->endpoint->uds_path != NULL)
 			sa = bogo_ip;
