@@ -56,6 +56,18 @@
 
 #include "VSC_vbe.h"
 
+struct vbp_state {
+	const char			*name;
+};
+
+#define VBP_STATE(n) static const struct vbp_state vbp_state_ ## n [1] = {{ .name = #n }}
+VBP_STATE(scheduled);
+VBP_STATE(running);
+VBP_STATE(cold);
+VBP_STATE(cooling);
+VBP_STATE(deleted);
+#undef VBP_STATE
+
 /* Default averaging rate, we want something pretty responsive */
 #define AVG_RATE			4
 
@@ -83,7 +95,7 @@ struct vbp_target {
 	double				rate;
 
 	vtim_real			due;
-	int				running;
+	const struct vbp_state		*state;
 	int				heap_idx;
 	struct pool_task		task[1];
 };
@@ -104,6 +116,8 @@ vbp_delete(struct vbp_target *vt)
 {
 	CHECK_OBJ_NOTNULL(vt, VBP_TARGET_MAGIC);
 
+	assert(vt->heap_idx == VBH_NOIDX);
+
 #define DN(x)	/**/
 	VRT_BACKEND_PROBE_HANDLE();
 #undef DN
@@ -177,6 +191,10 @@ VBP_Update_Backend(struct vbp_target *vt)
 	i = (vt->good < vt->threshold);
 	chg = (i != vt->backend->sick);
 	vt->backend->sick = i;
+	if (i && chg && (vt->backend->director != NULL &&
+	    VBE_is_ah_auto(vt->backend))) {
+		VBE_connwait_signal_all(vt->backend);
+	}
 
 	AN(vt->backend->vcl_name);
 	VSL(SLT_Backend_health, NO_VXID,
@@ -273,8 +291,9 @@ vbp_write_proxy_v1(struct vbp_target *vt, int *sock)
 static void
 vbp_poke(struct vbp_target *vt)
 {
-	int s, tmo, i, proxy_header, err;
+	int s, i, proxy_header, err;
 	vtim_real t_start, t_now, t_end;
+	vtim_dur tmo;
 	unsigned rlen, resp;
 	char buf[8192], *p;
 	struct pollfd pfda[1], *pfd = pfda;
@@ -304,11 +323,11 @@ vbp_poke(struct vbp_target *vt)
 		WRONG("Wrong probe protocol family");
 
 	t_now = VTIM_real();
-	tmo = (int)round((t_end - t_now) * 1e3);
+	tmo = t_end - t_now;
 	if (tmo <= 0) {
 		bprintf(vt->resp_buf,
 			"Open timeout %.3fs exceeded by %.3fs",
-			vt->timeout, t_now - t_end);
+			vt->timeout, -tmo);
 		VTCP_close(&s);
 		return;
 	}
@@ -347,15 +366,15 @@ vbp_poke(struct vbp_target *vt)
 		pfd->events = POLLIN;
 		pfd->revents = 0;
 		t_now = VTIM_real();
-		tmo = (int)round((t_end - t_now) * 1e3);
+		tmo = t_end - t_now;
 		if (tmo <= 0) {
 			bprintf(vt->resp_buf,
 			    "Poll timeout %.3fs exceeded by %.3fs",
-			    vt->timeout, t_now - t_end);
+			    vt->timeout, -tmo);
 			i = -1;
 			break;
 		}
-		i = poll(pfd, 1, tmo);
+		i = poll(pfd, 1, VTIM_poll_tmo(tmo));
 		if (i <= 0) {
 			if (!i) {
 				if (!vt->exp_close)
@@ -429,6 +448,33 @@ vbp_heap_insert(struct vbp_target *vt)
 /*--------------------------------------------------------------------
  */
 
+/*
+ * called when a task was successful or could not get scheduled
+ * returns non-NULL if target is to be deleted (outside mtx)
+ */
+static struct vbp_target *
+vbp_task_complete(struct vbp_target *vt)
+{
+	CHECK_OBJ_NOTNULL(vt, VBP_TARGET_MAGIC);
+
+	Lck_AssertHeld(&vbp_mtx);
+
+	assert(vt->heap_idx == VBH_NOIDX);
+
+	if (vt->state == vbp_state_running) {
+		vt->state = vbp_state_scheduled;
+		vt->due = VTIM_real() + vt->interval;
+		vbp_heap_insert(vt);
+		vt = NULL;
+	} else if (vt->state == vbp_state_cooling) {
+		vt->state = vbp_state_cold;
+		vt = NULL;
+	} else if (vt->state != vbp_state_deleted) {
+		WRONG(vt->state->name);
+	}
+	return (vt);
+}
+
 static void v_matchproto_(task_func_t)
 vbp_task(struct worker *wrk, void *priv)
 {
@@ -437,7 +483,6 @@ vbp_task(struct worker *wrk, void *priv)
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	CAST_OBJ_NOTNULL(vt, priv, VBP_TARGET_MAGIC);
 
-	AN(vt->running);
 	AN(vt->req);
 	assert(vt->req_len > 0);
 
@@ -447,25 +492,18 @@ vbp_task(struct worker *wrk, void *priv)
 	VBP_Update_Backend(vt);
 
 	Lck_Lock(&vbp_mtx);
-	if (vt->running < 0) {
-		assert(vt->heap_idx == VBH_NOIDX);
-		vbp_delete(vt);
-	} else {
-		vt->running = 0;
-		if (vt->heap_idx != VBH_NOIDX) {
-			vt->due = VTIM_real() + vt->interval;
-			VBH_delete(vbp_heap, vt->heap_idx);
-			vbp_heap_insert(vt);
-		}
-	}
+	vt = vbp_task_complete(vt);
 	Lck_Unlock(&vbp_mtx);
+	if (vt == NULL)
+		return;
+	vbp_delete(vt);
 }
 
 /*--------------------------------------------------------------------
  */
 
 static void * v_matchproto_(bgthread_t)
-vbp_thread(struct worker *wrk, void *priv)
+vbp_scheduler(struct worker *wrk, void *priv)
 {
 	vtim_real now, nxt;
 	struct vbp_target *vt;
@@ -482,22 +520,28 @@ vbp_thread(struct worker *wrk, void *priv)
 			(void)Lck_CondWaitUntil(&vbp_cond, &vbp_mtx, nxt);
 		} else if (vt->due > now) {
 			nxt = vt->due;
-			vt = NULL;
 			(void)Lck_CondWaitUntil(&vbp_cond, &vbp_mtx, nxt);
 		} else {
+			assert(vt->state == vbp_state_scheduled);
 			VBH_delete(vbp_heap, vt->heap_idx);
-			vt->due = now + vt->interval;
-			VBH_insert(vbp_heap, vt);
-			if (!vt->running) {
-				vt->running = 1;
-				vt->task->func = vbp_task;
-				vt->task->priv = vt;
-				Lck_Unlock(&vbp_mtx);
-				r = Pool_Task_Any(vt->task, TASK_QUEUE_REQ);
-				Lck_Lock(&vbp_mtx);
-				if (r)
-					vt->running = 0;
-			}
+			vt->state = vbp_state_running;
+			vt->task->func = vbp_task;
+			vt->task->priv = vt;
+			Lck_Unlock(&vbp_mtx);
+
+			r = Pool_Task_Any(vt->task, TASK_QUEUE_REQ);
+
+			Lck_Lock(&vbp_mtx);
+			if (r == 0)
+				continue;
+			vt = vbp_task_complete(vt);
+			if (vt == NULL)
+				continue;
+			Lck_Unlock(&vbp_mtx);
+
+			vbp_delete(vt);
+
+			Lck_Lock(&vbp_mtx);
 		}
 	}
 	NEEDLESS(Lck_Unlock(&vbp_mtx));
@@ -657,12 +701,24 @@ VBP_Control(const struct backend *be, int enable)
 
 	Lck_Lock(&vbp_mtx);
 	if (enable) {
-		assert(vt->heap_idx == VBH_NOIDX);
-		vt->due = VTIM_real();
-		vbp_heap_insert(vt);
+		if (vt->state == vbp_state_cooling) {
+			vt->state = vbp_state_running;
+		} else if (vt->state == vbp_state_cold) {
+			assert(vt->heap_idx == VBH_NOIDX);
+			vt->due = VTIM_real();
+			vbp_heap_insert(vt);
+			vt->state = vbp_state_scheduled;
+		} else
+			WRONG(vt->state->name);
 	} else {
-		assert(vt->heap_idx != VBH_NOIDX);
-		VBH_delete(vbp_heap, vt->heap_idx);
+		if (vt->state == vbp_state_running) {
+			vt->state = vbp_state_cooling;
+		} else if (vt->state == vbp_state_scheduled) {
+			assert(vt->heap_idx != VBH_NOIDX);
+			VBH_delete(vbp_heap, vt->heap_idx);
+			vt->state = vbp_state_cold;
+		} else
+			WRONG(vt->state->name);
 	}
 	Lck_Unlock(&vbp_mtx);
 }
@@ -685,6 +741,7 @@ VBP_Insert(struct backend *b, const struct vrt_backend_probe *vp,
 	ALLOC_OBJ(vt, VBP_TARGET_MAGIC);
 	XXXAN(vt);
 
+	vt->state = vbp_state_cold;
 	vt->conn_pool = tp;
 	VCP_AddRef(vt->conn_pool);
 	vt->backend = b;
@@ -709,19 +766,14 @@ VBP_Remove(struct backend *be)
 	be->sick = 1;
 	be->probe = NULL;
 	vt->backend = NULL;
-	if (vt->running) {
-		// task scheduled, it calls vbp_delete()
-		vt->running = -1;
+	if (vt->state == vbp_state_cooling) {
+		vt->state = vbp_state_deleted;
 		vt = NULL;
-	} else if (vt->heap_idx != VBH_NOIDX) {
-		// task done, not yet rescheduled
-		VBH_delete(vbp_heap, vt->heap_idx);
-	}
+	} else
+		assert(vt->state == vbp_state_cold);
 	Lck_Unlock(&vbp_mtx);
-	if (vt != NULL) {
-		assert(vt->heap_idx == VBH_NOIDX);
+	if (vt != NULL)
 		vbp_delete(vt);
-	}
 }
 
 /*-------------------------------------------------------------------*/
@@ -730,18 +782,11 @@ static int v_matchproto_(vbh_cmp_t)
 vbp_cmp(void *priv, const void *a, const void *b)
 {
 	const struct vbp_target *aa, *bb;
-	int ar, br;
 
 	AZ(priv);
 	CAST_OBJ_NOTNULL(aa, a, VBP_TARGET_MAGIC);
 	CAST_OBJ_NOTNULL(bb, b, VBP_TARGET_MAGIC);
 
-	ar = aa->running == 0;
-	br = bb->running == 0;
-
-	if (ar != br)
-		return (ar);
-
 	return (aa->due < bb->due);
 }
 
@@ -766,5 +811,5 @@ VBP_Init(void)
 	vbp_heap = VBH_new(NULL, vbp_cmp, vbp_update);
 	AN(vbp_heap);
 	PTOK(pthread_cond_init(&vbp_cond, NULL));
-	WRK_BgThread(&thr, "backend-poller", vbp_thread, NULL);
+	WRK_BgThread(&thr, "backend-probe-scheduler", vbp_scheduler, NULL);
 }
