@@ -59,22 +59,47 @@ VDP_Panic(struct vsb *vsb, const struct vdp_ctx *vdc)
 	VSB_cat(vsb, "},\n");
 }
 
-
+/*
+ * Ensure that transports have called VDP_Close()
+ * to avoid leaks in VDPs
+ */
+void
+VDP_Fini(const struct vdp_ctx *vdc)
+{
+	assert(VTAILQ_EMPTY(&vdc->vdp));
+}
 
 void
 VDP_Init(struct vdp_ctx *vdc, struct worker *wrk, struct vsl_log *vsl,
-    struct req *req)
+    const struct req *req, const struct busyobj *bo, intmax_t *clen)
 {
 	AN(vdc);
 	CHECK_OBJ_NOTNULL(wrk, WORKER_MAGIC);
 	AN(vsl);
-	CHECK_OBJ_NOTNULL(req, REQ_MAGIC);
+
+	AN(clen);
+
+	assert((req ? 1 : 0) ^ (bo ? 1 : 0));
+
+	AN(clen);
+	assert(*clen >= -1);
 
 	INIT_OBJ(vdc, VDP_CTX_MAGIC);
 	VTAILQ_INIT(&vdc->vdp);
 	vdc->wrk = wrk;
 	vdc->vsl = vsl;
-	vdc->req = req;
+	vdc->clen = clen;
+
+	if (req != NULL) {
+		CHECK_OBJ(req, REQ_MAGIC);
+		vdc->oc = req->objcore;
+		vdc->hp = req->resp;
+	}
+	else {
+		CHECK_OBJ_NOTNULL(bo, BUSYOBJ_MAGIC);
+		vdc->oc = bo->bereq_body;
+		vdc->hp = bo->bereq;
+	}
 }
 
 /* VDP_bytes
@@ -144,10 +169,13 @@ VDP_Push(VRT_CTX, struct vdp_ctx *vdc, struct ws *ws, const struct vdp *vdp,
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
+	CHECK_OBJ_ORNULL(vdc->oc, OBJCORE_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc->hp, HTTP_MAGIC);
+	AN(vdc->clen);
+	assert(*vdc->clen >= -1);
 	AN(ws);
 	AN(vdp);
 	AN(vdp->name);
-	AN(vdp->bytes);
 
 	if (vdc->retval)
 		return (vdc->retval);
@@ -168,17 +196,17 @@ VDP_Push(VRT_CTX, struct vdp_ctx *vdc, struct ws *ws, const struct vdp *vdp,
 	vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
 
 	AZ(vdc->retval);
-	if (vdpe->vdp->init == NULL)
-		return (vdc->retval);
-
-	vdc->retval = vdpe->vdp->init(ctx, vdc, &vdpe->priv,
-	    vdpe == vdc->nxt ? vdc->req->objcore : NULL);
+	if (vdpe->vdp->init != NULL)
+		vdc->retval = vdpe->vdp->init(ctx, vdc, &vdpe->priv);
+	vdc->oc = NULL;
 
 	if (vdc->retval > 0) {
 		VTAILQ_REMOVE(&vdc->vdp, vdpe, list);
 		vdc->nxt = VTAILQ_FIRST(&vdc->vdp);
 		vdc->retval = 0;
 	}
+	else if (vdc->retval == 0)
+		AN(vdp->bytes);
 	return (vdc->retval);
 }
 
@@ -246,7 +274,9 @@ VDP_DeliverObj(struct vdp_ctx *vdc, struct objcore *oc)
 	CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
 	CHECK_OBJ_NOTNULL(vdc->wrk, WORKER_MAGIC);
 	AN(vdc->vsl);
-	vdc->req = NULL;
+	AZ(vdc->oc);
+	vdc->hp = NULL;
+	vdc->clen = NULL;
 	final = oc->flags & (OC_F_PRIVATE | OC_F_HFM | OC_F_HFP) ? 1 : 0;
 	r = ObjIterate(vdc->wrk, oc, vdc, vdp_objiterate, final);
 	if (r < 0)
