@@ -31,12 +31,6 @@
 
 #include "config.h"
 
-#ifdef WITH_UNWIND
-#  include <libunwind.h>
-#else
-#  include <execinfo.h>
-#endif
-
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
@@ -51,6 +45,7 @@
 #include "storage/storage.h"
 #include "vcli_serve.h"
 #include "vtim.h"
+#include "vbt.h"
 #include "vcs.h"
 #include "vtcp.h"
 #include "vsa.h"
@@ -67,7 +62,6 @@
 
 static struct vsb pan_vsb_storage, *pan_vsb;
 static pthread_mutex_t panicstr_mtx;
-static pthread_t panicy;
 
 static void pan_sess(struct vsb *, const struct sess *);
 static void pan_req(struct vsb *, const struct req *);
@@ -379,8 +373,10 @@ pan_busyobj(struct vsb *vsb, const struct busyobj *bo)
 
 	if (bo->vfc != NULL)
 		pan_vfp(vsb, bo->vfc);
-	if (bo->vfp_filter_list != NULL)
-		VSB_printf(vsb, "filter_list = \"%s\",\n", bo->vfp_filter_list);
+	if (bo->vfp_filter_list != NULL) {
+		VSB_printf(vsb, "vfp_filter_list = \"%s\",\n",
+		    bo->vfp_filter_list);
+	}
 
 	WS_Panic(vsb, bo->ws);
 	VSB_printf(vsb, "ws_bo = %p,\n", (void *)bo->ws_bo);
@@ -603,93 +599,16 @@ pan_sess(struct vsb *vsb, const struct sess *sp)
 
 /*--------------------------------------------------------------------*/
 
-#ifdef WITH_UNWIND
-
 static void
 pan_backtrace(struct vsb *vsb)
 {
-	unw_cursor_t cursor; unw_context_t uc;
-	unw_word_t ip, sp;
-	unw_word_t offp;
-	char fname[1024];
-	int ret;
 
 	VSB_cat(vsb, "Backtrace:\n");
 	VSB_indent(vsb, 2);
-
-	ret = unw_getcontext(&uc);
-	if (ret != 0) {
-		VSB_printf(vsb, "Backtrace not available "
-		    "(unw_getcontext returned %d)\n", ret);
-		return;
-	}
-	ret = unw_init_local(&cursor, &uc);
-	if (ret != 0) {
-		VSB_printf(vsb, "Backtrace not available "
-		    "(unw_init_local returned %d)\n", ret);
-		return;
-	}
-	while (unw_step(&cursor) > 0) {
-		fname[0] = '\0';
-		if (!unw_get_reg(&cursor, UNW_REG_IP, &ip))
-			VSB_printf(vsb, "ip=0x%lx", (long) ip);
-		if (!unw_get_reg(&cursor, UNW_REG_SP, &sp))
-			VSB_printf(vsb, " sp=0x%lx", (long) sp);
-		if (!unw_get_proc_name(&cursor, fname, sizeof(fname), &offp))
-			VSB_printf(vsb, " <%s+0x%lx>",
-			    fname[0] ? fname : "<unknown>", (long)offp);
-		VSB_putc(vsb, '\n');
-	}
-
+	VBT_format(vsb);
 	VSB_indent(vsb, -2);
 }
 
-#else /* WITH_UNWIND */
-
-#define BACKTRACE_LEVELS	20
-
-static void
-pan_backtrace(struct vsb *vsb)
-{
-	void *array[BACKTRACE_LEVELS];
-	size_t size;
-	size_t i;
-	char **strings;
-	char *p;
-	char buf[32];
-
-	size = backtrace (array, BACKTRACE_LEVELS);
-	if (size > BACKTRACE_LEVELS) {
-		VSB_printf(vsb, "Backtrace not available (ret=%zu)\n", size);
-		return;
-	}
-	VSB_cat(vsb, "Backtrace:\n");
-	VSB_indent(vsb, 2);
-	for (i = 0; i < size; i++) {
-		bprintf(buf, "%p", array[i]);
-		VSB_printf(vsb, "%s: ", buf);
-		strings = backtrace_symbols(&array[i], 1);
-		if (strings == NULL || strings[0] == NULL) {
-			VSB_cat(vsb, "(?)");
-		} else {
-			p = strings[0];
-			if (!memcmp(buf, p, strlen(buf))) {
-				p += strlen(buf);
-				if (*p == ':')
-					p++;
-				while (*p == ' ')
-					p++;
-			}
-			VSB_printf(vsb, "%s", p);
-		}
-		VSB_cat(vsb, "\n");
-		free(strings);
-	}
-	VSB_indent(vsb, -2);
-}
-
-#endif /* WITH_UNWIND */
-
 #ifdef HAVE_PTHREAD_GETATTR_NP
 static void
 pan_threadattr(struct vsb *vsb)
@@ -744,18 +663,23 @@ pan_ic(const char *func, const char *file, int line, const char *cond,
 	struct busyobj *bo;
 	struct worker *wrk;
 	struct sigaction sa;
-	int err = errno;
+	int i, err = errno;
 
-	/* If we already panicing in another thread, do nothing */
-	while (heritage.panic_str[0] && panicy != pthread_self())
-		sleep(1);
-
-	if (pthread_mutex_lock(&panicstr_mtx)) {
-		/* Reentrant panic */
+	if (pthread_getspecific(panic_key) != NULL) {
 		VSB_cat(pan_vsb, "\n\nPANIC REENTRANCY\n\n");
 		abort();
 	}
-	panicy = pthread_self();
+
+	/* If we already panicking in another thread, do nothing */
+	do {
+		i = pthread_mutex_trylock(&panicstr_mtx);
+		if (i != 0)
+			sleep (1);
+	} while (i != 0);
+
+	assert (VSB_len(pan_vsb) == 0);
+
+	AZ(pthread_setspecific(panic_key, pan_vsb));
 
 	/*
 	 * should we trigger a SIGSEGV while handling a panic, our sigsegv
@@ -780,7 +704,7 @@ pan_ic(const char *func, const char *file, int line, const char *cond,
 		break;
 	case VAS_MISSING:
 		VSB_printf(pan_vsb,
-		    "Missing errorhandling code in %s(), %s line %d:\n"
+		    "Missing error handling code in %s(), %s line %d:\n"
 		    "  Condition(%s) not true.\n",
 		    func, file, line, cond);
 		break;
@@ -844,6 +768,14 @@ pan_ic(const char *func, const char *file, int line, const char *cond,
 	VSB_putc(pan_vsb, '\0');	/* NUL termination */
 
 	v_gcov_flush();
+
+	/*
+	 * Do a little song and dance for static checkers which
+	 * are not smart enough to figure out that calling abort()
+	 * with a mutex held is OK and probably very intentional.
+	 */
+	if (pthread_getspecific(panic_key))	/* ie: always */
+		abort();
 	PTOK(pthread_mutex_unlock(&panicstr_mtx));
 	abort();
 }
