@@ -73,14 +73,15 @@ V2D_Init(void)
 /**********************************************************************/
 
 static int v_matchproto_(vdp_init_f)
-h2_init(VRT_CTX, struct vdp_ctx *vdc, void **priv, struct objcore *oc)
+h2_init(VRT_CTX, struct vdp_ctx *vdc, void **priv)
 {
+	struct h2_req *r2;
 
 	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
 	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
-	(void)oc;
-	CHECK_OBJ_NOTNULL(vdc->req, REQ_MAGIC);
-	*priv = vdc->req->transport_priv;
+	AN(priv);
+	CAST_OBJ_NOTNULL(r2, *priv, H2_REQ_MAGIC);
+	(void)r2;
 	return (0);
 }
 
@@ -111,12 +112,12 @@ h2_fini(struct vdp_ctx *vdc, void **priv)
 }
 
 static int v_matchproto_(vdp_bytes_f)
-h2_bytes(struct vdp_ctx *vdx, enum vdp_action act, void **priv,
+h2_bytes(struct vdp_ctx *vdc, enum vdp_action act, void **priv,
     const void *ptr, ssize_t len)
 {
 	struct h2_req *r2;
 
-	CHECK_OBJ_NOTNULL(vdx, VDP_CTX_MAGIC);
+	CHECK_OBJ_NOTNULL(vdc, VDP_CTX_MAGIC);
 	CAST_OBJ_NOTNULL(r2, *priv, H2_REQ_MAGIC);
 	(void)act;
 
@@ -124,9 +125,9 @@ h2_bytes(struct vdp_ctx *vdx, enum vdp_action act, void **priv,
 		return (-1);
 	if (len == 0)
 		return (0);
-	H2_Send_Get(vdx->wrk, r2->h2sess, r2);
-	vdx->bytes_done = 0;
-	H2_Send(vdx->wrk, r2, H2_F_DATA, H2FF_NONE, len, ptr, &vdx->bytes_done);
+	H2_Send_Get(vdc->wrk, r2->h2sess, r2);
+	vdc->bytes_done = 0;
+	H2_Send(vdc->wrk, r2, H2_F_DATA, H2FF_NONE, len, ptr, &vdc->bytes_done);
 	H2_Send_Rel(r2->h2sess, r2);
 	return (0);
 }
@@ -346,7 +347,7 @@ h2_deliver(struct req *req, struct boc *boc, int sendbody)
 	if (sendbody) {
 		INIT_OBJ(ctx, VRT_CTX_MAGIC);
 		VCL_Req2Ctx(ctx, req);
-		if (!VDP_Push(ctx, req->vdc, req->ws, &h2_vdp, NULL))
+		if (! VDP_Push(ctx, req->vdc, req->ws, &h2_vdp, r2))
 			(void)VDP_DeliverObj(req->vdc, req->objcore);
 	}
 
