@@ -62,6 +62,7 @@
 #include "vtim.h"
 #include "waiter/mgt_waiter.h"
 #include "vsa.h"
+#include "vus.h"
 
 struct heritage		heritage;
 unsigned		d_flag = 0;
@@ -78,23 +79,43 @@ static const char opt_spec[] = "?a:b:CdE:f:Fh:i:I:j:l:M:n:P:p:r:S:s:T:t:VW:x:";
 
 /*--------------------------------------------------------------------*/
 
+// Basic options documentation with compile time dependencies
+static void
+mgt_DumpOptions(void)
+{
+	printf(".. _opt_n:\n\n");
+	printf("-n workdir\n\n");
+	printf("  Runtime directory for the shared memory, "
+	       "compiled VCLs etc.\n\n");
+	printf("  In performance critical applications, this directory "
+	       "should be on a RAM backed filesystem.\n\n");
+	VIN_DumpDefaults();
+}
+
 static void
 usage(void)
 {
-#define FMT "  %-28s # %s\n"
+	char *p;
+
+#define FMT_NONL "  %-35s # %s"
+#define FMT FMT_NONL "\n"
 
 	printf( "Usage: varnishd [options]\n");
 
 	printf("\nBasic options:\n");
 
-	printf(FMT, "-a [<name>=]address[:port][,proto]",
-	    "HTTP listen address and port");
-	printf(FMT, "   [,user=<u>][,group=<g>]",
+	printf(FMT, "[-a [name=][listen_address",
+	    "HTTP listen address, protocol, options.");
+	printf(FMT, "     [,PROTO|,option=value,...]]",
 	    "Can be specified multiple times.");
-	printf(FMT, "   [,mode=<m>]", "  default: \":80,HTTP\"");
-	printf(FMT, "", "Proto can be \"PROXY\" or \"HTTP\" (default)");
-	printf(FMT, "", "user, group and mode set permissions for");
-	printf(FMT, "", "  a Unix domain socket.");
+	printf(FMT, "",
+	    "  default: \":80,HTTP\"");
+	printf(FMT, "  options:",
+	    "Proto can be \"PROXY\" or \"HTTP\" (default)");
+	printf(FMT, "    [,user=<u>][,group=<g>]",
+	    "user, group and mode set permissions for");
+	printf(FMT, "    [,mode=<m>]",
+	    "  a Unix domain socket.");
 	printf(FMT, "-b none", "No backend");
 	printf(FMT, "-b [addr[:port]|path]", "Backend address and port");
 	printf(FMT, "", "  or socket file path");
@@ -103,6 +124,11 @@ usage(void)
 	printf(FMT, "", "Can be specified multiple times.");
 	printf(FMT, "-n dir", "Working directory");
 
+	p = VIN_n_Arg(NULL);
+	AN(p);
+	printf(FMT_NONL "  default: %s\n", "", "", p);
+	free(p);
+
 	printf("\n-b can be used only once, and not together with -f\n");
 
 	printf("\nDocumentation options:\n");
@@ -112,6 +138,7 @@ usage(void)
 	printf(FMT, "-x cli", "CLI command documentation");
 	printf(FMT, "-x builtin", "Builtin VCL program");
 	printf(FMT, "-x optstring", "List of getopt options");
+	printf(FMT, "-x options", "Dynamic options documentation");
 
 	printf("\nOperations options:\n");
 
@@ -153,6 +180,9 @@ usage(void)
 	printf(FMT, "-j jail[,options]", "Jail specification");
 #ifdef HAVE_SETPPRIV
 	printf(FMT, "", "  -j solaris");
+#endif
+#ifdef __linux__
+	printf(FMT, "", "  -j linux");
 #endif
 	printf(FMT, "", "  -j unix");
 	printf(FMT, "", "  -j none");
@@ -263,10 +293,16 @@ make_secret(const char *dirname)
 	assert(asprintf(&fn, "%s/_.secret", dirname) > 0);
 
 	VJ_master(JAIL_MASTER_FILE);
-	fdo = open(fn, O_RDWR|O_CREAT|O_TRUNC, 0640);
-	if (fdo < 0)
+	if (unlink(fn) < 0 && errno != ENOENT) {
+		ARGV_ERR("Cannot remove preexisting secret-file in %s (%s)\n",
+		    dirname, VAS_errtxt(errno));
+	}
+
+	fdo = open(fn, O_RDWR|O_CREAT|O_EXCL, 0640);
+	if (fdo < 0) {
 		ARGV_ERR("Cannot create secret-file in %s (%s)\n",
 		    dirname, VAS_errtxt(errno));
+	}
 
 	for (i = 0; i < 256; i++) {
 		AZ(VRND_RandomCrypto(&b, 1));
@@ -285,9 +321,12 @@ mgt_Cflag_atexit(void)
 	/* Only master process */
 	if (getpid() != heritage.mgt_pid)
 		return;
-	vext_cleanup(1);
+	if (arg_list_count("E")) {
+		vext_cleanup(1);
+		VJ_rmdir("vext_cache");
+	}
 	VJ_rmdir("vmod_cache");
-	VJ_rmdir("vext_cache");
+	VJ_rmdir("worker_tmpdir");
 	(void)chdir("/");
 	VJ_rmdir(workdir);
 }
@@ -342,6 +381,8 @@ mgt_x_arg(const char *x_arg)
 		mgt_DumpBuiltin();
 	else if (!strcmp(x_arg, "optstring"))
 		(void)printf("%s\n", opt_spec);
+	else if (!strcmp(x_arg, "options"))
+		mgt_DumpOptions();
 	else
 		ARGV_ERR("Invalid -x argument\n");
 }
@@ -483,10 +524,10 @@ mgt_b_conv(const char *b_arg)
 	VSB_cat(vsb, "backend default ");
 	if (!strcasecmp(b_arg, "none"))
 		VSB_cat(vsb, "none;\n");
-	else if (*b_arg != '/')
-		VSB_printf(vsb, "{\n    .host = \"%s\";\n}\n", b_arg);
-	else
+	else if (VUS_is(b_arg))
 		VSB_printf(vsb, "{\n    .path = \"%s\";\n}\n", b_arg);
+	else
+		VSB_printf(vsb, "{\n    .host = \"%s\";\n}\n", b_arg);
 	AZ(VSB_finish(vsb));
 	fa->src = strdup(VSB_data(vsb));
 	AN(fa->src);
@@ -581,7 +622,7 @@ main(int argc, char * const *argv)
 	const char *i_arg = NULL;
 	const char *j_arg = NULL;
 	const char *h_arg = "critbit";
-	const char *n_arg = NULL;
+	const char *n_arg = getenv("VARNISH_DEFAULT_N");
 	const char *S_arg = NULL;
 	const char *s_arg = "default,100m";
 	const char *W_arg = NULL;
@@ -832,8 +873,7 @@ main(int argc, char * const *argv)
 		    workdir, VAS_errtxt(errno));
 
 	VJ_master(JAIL_MASTER_SYSTEM);
-	AZ(system("rm -rf vmod_cache"));
-	AZ(system("rm -rf vext_cache"));
+	AZ(system("rm -rf vmod_cache vext_cache worker_tmpdir"));
 	VJ_master(JAIL_MASTER_LOW);
 
 	if (VJ_make_subdir("vmod_cache", "VMOD cache", NULL)) {
@@ -843,12 +883,25 @@ main(int argc, char * const *argv)
 	}
 
 	if (arg_list_count("E") &&
-	    VJ_make_subdir("vext_cache", "VMOD cache", NULL)) {
+	    VJ_make_subdir("vext_cache", "VEXT cache", NULL)) {
 		ARGV_ERR(
 		    "Cannot create vmod directory (%s/vext_cache): %s\n",
 		    workdir, VAS_errtxt(errno));
 	}
 
+	if (VJ_make_subdir("worker_tmpdir",
+	    "TMPDIR for the worker process", NULL)) {
+		ARGV_ERR(
+		    "Cannot create vmod directory (%s/worker_tmpdir): %s\n",
+		    workdir, VAS_errtxt(errno));
+	}
+
+	o = open("worker_tmpdir", O_RDONLY);
+	VJ_master(JAIL_MASTER_SYSTEM);
+	VJ_fix_fd(o, JAIL_FIXFD_WRKTMP);
+	VJ_master(JAIL_MASTER_LOW);
+	closefd(&o);
+
 	if (C_flag)
 		AZ(atexit(mgt_Cflag_atexit));
 
@@ -876,7 +929,7 @@ main(int argc, char * const *argv)
 			alp->priv = create_pid_file(&pid, "%s", alp->val);
 	}
 
-	/* Implict -P argument */
+	/* Implicit -P argument */
 	alp = arg_list_add('P', NULL);
 	alp->priv = create_pid_file(&pid, "%s/_.pid", workdir);
 
