@@ -93,7 +93,7 @@ enum e_frag {
 };
 
 struct fragment {
-	unsigned		gen;
+	uint64_t		gen;
 	const char		*b, *e;
 };
 
@@ -144,7 +144,7 @@ static struct ctx {
 
 	FILE			*fo;
 	struct vsb		*vsb;
-	unsigned		gen;
+	uint64_t		gen;
 	VTAILQ_HEAD(,format)	format;
 	int			quote_how;
 	char			*missing_string;
@@ -162,6 +162,8 @@ static struct ctx {
 	int64_t			vxid;
 } CTX;
 
+static void parse_format(const char *format);
+
 static void
 openout(int append)
 {
@@ -681,6 +683,10 @@ parse_x_format(char *buf)
 		addf_vsl((enum VSL_tag_e)slt, lval, r);
 		return;
 	}
+	if (!strcmp(buf, "Varnish:default_format")) {
+		parse_format(FORMAT);
+		return;
+	}
 	VUT_Error(vut, 1, "Unknown formatting extension: %s", buf);
 }
 
