@@ -55,7 +55,6 @@
 #define HAPROXY_OPT_DAEMON	"-D"
 #define HAPROXY_SIGNAL		SIGINT
 #define HAPROXY_EXPECT_EXIT	(128 + HAPROXY_SIGNAL)
-#define HAPROXY_GOOD_CONF	"Configuration file is valid"
 
 struct envar {
 	VTAILQ_ENTRY(envar) list;
@@ -268,7 +267,7 @@ haproxy_cli_recv(struct haproxy_cli *hc)
 		}
 		/* Connection closed. */
 		if (ret == 0) {
-			if (hc->rxbuf[rdz - 1] != '\n')
+			if (rdz > 0 && hc->rxbuf[rdz - 1] != '\n')
 				vtc_fatal(hc->vl,
 				    "CLI rx timeout (fd: %d %.3fs ret: %zd)",
 				    hc->sock, hc->timeout, ret);
@@ -291,7 +290,7 @@ haproxy_cli_recv(struct haproxy_cli *hc)
  * SECTION: haproxy.cli.expect
  * expect OP STRING
  *         Regex match the CLI reception buffer with STRING
- *         if OP is ~ or, on the contraty, if OP is !~ check that there is
+ *         if OP is ~ or, on the contrary, if OP is !~ check that there is
  *         no regex match.
  */
 static void v_matchproto_(cmd_f)
@@ -1077,8 +1076,9 @@ cmd_haproxy(CMD_ARGS)
 		if (!strcmp(*av, "-conf-OK")) {
 			AN(av[1]);
 			haproxy_store_conf(h, av[1], 0);
+			h->expect_exit = 0;
+			haproxy_check_conf(h, "");
 			av++;
-			haproxy_check_conf(h, HAPROXY_GOOD_CONF);
 			continue;
 		}
 		if (!strcmp(*av, "-conf-BAD")) {
