@@ -373,6 +373,13 @@ cmd_http_expect(CMD_ARGS)
 	vtc_expect(vl, av[0], lhs, cmp, av[2], rhs);
 }
 
+/* SECTION: client-server.spec.expect_pattern
+ *
+ * expect_pattern
+ *
+ * Expect as the http body the test pattern generated by chunkedlen ('0'..'7'
+ * repeating).
+ */
 static void
 cmd_http_expect_pattern(CMD_ARGS)
 {
@@ -764,12 +771,12 @@ cmd_http_gunzip(CMD_ARGS)
 }
 
 /**********************************************************************
- * Handle common arguments of a transmited request or response
+ * Handle common arguments of a transmitted request or response
  */
 
 static char* const *
 http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
-    char *body, unsigned nohost, unsigned nodate)
+    char *body, unsigned nohost, unsigned nodate, unsigned noserver, unsigned nouseragent)
 {
 	long bodylen = 0;
 	char *b, *c;
@@ -795,6 +802,10 @@ http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
 				nohost = 1;
 			if (!strncasecmp(av[1], "Date:", 5))
 				nodate = 1;
+			if (!strncasecmp(av[1], "Server:", 7))
+				noserver = 1;
+			if (!strncasecmp(av[1], "User-Agent:", 11))
+				nouseragent = 1;
 			VSB_printf(hp->vsb, "%s%s", av[1], nl);
 			av++;
 		} else if (!strcmp(*av, "-hdrlen")) {
@@ -839,28 +850,13 @@ http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
 			body = synth_body(av[1], 0);
 			bodylen = strlen(body);
 			av++;
-		} else if (!strcmp(*av, "-gzipresidual")) {
-			hp->gzipresidual = strtoul(av[1], NULL, 0);
-			av++;
-		} else if (!strcmp(*av, "-gziplevel")) {
-			hp->gziplevel = strtoul(av[1], NULL, 0);
-			av++;
-		} else if (!strcmp(*av, "-gziplen")) {
-			assert(body == nullbody);
-			free(body);
-			b = synth_body(av[1], 1);
-			vtc_gzip(hp, b, &body, &bodylen);
-			free(b);
-			VSB_printf(hp->vsb, "Content-Encoding: gzip%s", nl);
-			// vtc_hexdump(hp->vl, 4, "gzip", (void*)body, bodylen);
-			av++;
-		} else if (!strcmp(*av, "-gzipbody")) {
-			assert(body == nullbody);
-			free(body);
-			vtc_gzip(hp, av[1], &body, &bodylen);
-			VSB_printf(hp->vsb, "Content-Encoding: gzip%s", nl);
-			// vtc_hexdump(hp->vl, 4, "gzip", (void*)body, bodylen);
-			av++;
+		} else if (!strncmp(*av, "-gzip", 5)) {
+			l = vtc_gzip_cmd(hp, av, &body, &bodylen);
+			if (l == 0)
+				break;
+			av += l;
+			if (l > 1)
+				VSB_printf(hp->vsb, "Content-Encoding: gzip%s", nl);
 		} else
 			break;
 	}
@@ -874,6 +870,10 @@ http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
 		macro_cat(vl, hp->vsb, "date", NULL);
 		VSB_cat(hp->vsb, nl);
 	}
+	if (!noserver)
+		VSB_printf(hp->vsb, "Server: %s%s", hp->sess->name, nl);
+	if (!nouseragent)
+		VSB_printf(hp->vsb, "User-Agent: %s%s", hp->sess->name, nl);
 	if (body != NULL && !nolen)
 		VSB_printf(hp->vsb, "Content-Length: %ld%s", bodylen, nl);
 	VSB_cat(hp->vsb, nl);
@@ -893,7 +893,7 @@ http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
  *
  *         The only thing different between a request and a response, apart
  *         from who can send them is that the first line (request line vs
- *         status line), so all the options are prety much the same.
+ *         status line), so all the options are pretty much the same.
  *
  *         \-method STRING (txreq only)
  *                 What method to use (default: "GET").
@@ -914,6 +914,12 @@ http_tx_parse_args(char * const *av, struct vtclog *vl, struct http *hp,
  *         \-reason STRING (txresp only)
  *                 What message to put in the status line (default: "OK").
  *
+ *         \-noserver (txresp only)
+ *                 Don't include a Server header with the id of the server.
+ *
+ *         \-nouseragent (txreq only)
+ *                 Don't include a User-Agent header with the id of the client.
+ *
  *         These three switches can appear in any order but must come before the
  *         following ones.
  *
@@ -975,6 +981,7 @@ cmd_http_txresp(CMD_ARGS)
 	const char *status = "200";
 	const char *reason = "OK";
 	char* body = NULL;
+	unsigned noserver = 0;
 
 	(void)vl;
 	CAST_OBJ_NOTNULL(hp, priv, HTTP_MAGIC);
@@ -995,6 +1002,9 @@ cmd_http_txresp(CMD_ARGS)
 			reason = av[1];
 			av++;
 			continue;
+		} else if (!strcmp(*av, "-noserver")) {
+			noserver = 1;
+			continue;
 		} else
 			break;
 	}
@@ -1004,7 +1014,7 @@ cmd_http_txresp(CMD_ARGS)
 	/* send a "Content-Length: 0" header unless something else happens */
 	REPLACE(body, "");
 
-	av = http_tx_parse_args(av, vl, hp, body, 1, 0);
+	av = http_tx_parse_args(av, vl, hp, body, 1, 0, noserver, 1);
 	if (*av != NULL)
 		vtc_fatal(hp->vl, "Unknown http txresp spec: %s\n", *av);
 
@@ -1200,6 +1210,7 @@ cmd_http_txreq(CMD_ARGS)
 	const char *proto = "HTTP/1.1";
 	const char *up = NULL;
 	unsigned nohost;
+	unsigned nouseragent = 0;
 
 	(void)vl;
 	CAST_OBJ_NOTNULL(hp, priv, HTTP_MAGIC);
@@ -1225,6 +1236,8 @@ cmd_http_txreq(CMD_ARGS)
 		} else if (!hp->sfd && !strcmp(*av, "-up")) {
 			up = av[1];
 			av++;
+		} else if (!strcmp(*av, "-nouseragent")) {
+			nouseragent = 1;
 		} else
 			break;
 	}
@@ -1236,7 +1249,7 @@ cmd_http_txreq(CMD_ARGS)
 				"HTTP2-Settings: %s%s", nl, nl, up, nl);
 
 	nohost = strcmp(proto, "HTTP/1.1") != 0;
-	av = http_tx_parse_args(av, vl, hp, NULL, nohost, 1);
+	av = http_tx_parse_args(av, vl, hp, NULL, nohost, 1, 1, nouseragent);
 	if (*av != NULL)
 		vtc_fatal(hp->vl, "Unknown http txreq spec: %s\n", *av);
 	http_write(hp, 4, "txreq");
@@ -1571,6 +1584,53 @@ cmd_http_accept(CMD_ARGS)
 	vtc_log(vl, 3, "Accepted socket fd is %d", hp->sess->fd);
 }
 
+/* SECTION: client-server.spec.shutdown
+ *
+ * shutdown
+ *	Initiate shutdown.
+ *
+ *	\-read
+ *		Shutdown the read direction.
+ *	\-write
+ *		Shutdown the write direction.
+ *
+ *	The default is both direction.
+ */
+static void
+cmd_http_shutdown(CMD_ARGS)
+{
+	struct http *hp;
+	int how = SHUT_RDWR;
+	const char *str[] = {
+		[SHUT_RD]	= "RD",
+		[SHUT_WR]	= "WR",
+		[SHUT_RDWR]	= "RDWR",
+	};
+
+	(void)vl;
+	CAST_OBJ_NOTNULL(hp, priv, HTTP_MAGIC);
+	AZ(strcmp(av[0], "shutdown"));
+	av++;
+
+	if (*av != NULL) {
+		if (!strcmp(*av, "-read")) {
+			how = SHUT_RD;
+			av++;
+		} else if (!strcmp(*av, "-write")) {
+			how = SHUT_WR;
+			av++;
+		}
+	}
+
+	if (*av != NULL)
+		vtc_fatal(hp->vl, "Unknown http shutdown spec: %s\n", *av);
+
+	vtc_log(vl, 4, "Shutting down fd (%s): %d", str[how], hp->sess->fd);
+	if (shutdown(hp->sess->fd, how) < 0)
+		vtc_log(vl, hp->fatal, "Shutdown failed: %s", strerror(errno));
+	vtc_log(vl, 3, "Shutdown socket fd (%d): %d", how, hp->sess->fd);
+}
+
 /* SECTION: client-server.spec.fatal
  *
  * fatal|non_fatal
@@ -1749,6 +1809,7 @@ const struct cmds http_cmds[] = {
 	CMD_HTTP(send_n)
 	CMD_HTTP(send_urgent)
 	CMD_HTTP(sendhex)
+	CMD_HTTP(shutdown)
 	CMD_HTTP(timeout)
 
 	/* spec */
