@@ -52,7 +52,7 @@
 #define BUF_SIZE	(1024*2048)
 
 static const char *const h2_errs[] = {
-#define H2_ERROR(n,v,sc,r,t) [v] = #n,
+#define H2_ERROR(n,v,sc,g,r,t) [v] = #n,
 #include <tbl/h2_error.h>
 	NULL
 };
@@ -169,7 +169,7 @@ http_write(const struct http *hp, int lvl,
 }
 
 static int
-get_bytes(const struct http *hp, char *buf, int n)
+get_bytes(const struct http *hp, char *buf, size_t n)
 {
 	int i;
 	struct pollfd pfd[1];
@@ -197,7 +197,7 @@ get_bytes(const struct http *hp, char *buf, int n)
 		i = read(hp->sess->fd, buf, n);
 		if (!(pfd[0].revents & POLLIN))
 			vtc_log(hp->vl, 4,
-			    "HTTP2 rx poll (fd:%d revents: %x n=%d, i=%d)",
+			    "HTTP2 rx poll (fd:%d revents: %x n=%zu, i=%d)",
 			    hp->sess->fd, pfd[0].revents, n, i);
 		if (i == 0)
 			vtc_log(hp->vl, 3,
@@ -467,22 +467,20 @@ decode_hdr(struct http *hp, struct hpk_hdr *h, const struct vsb *vsb)
 		r = HPK_DecHdr(iter, h + n);
 		if (r == hpk_err )
 			break;
-		vtc_log(hp->vl, 4,
-				"header[%2d]: %s : %s",
-				n,
-				h[n].key.ptr,
-				h[n].value.ptr);
+		vtc_log(hp->vl, 4, "header[%2d]: %s: %s",
+		    n, h[n].key.ptr, h[n].value.ptr);
 		n++;
 		if (r == hpk_done)
 			break;
 	}
 
-	if (r != hpk_done)
+	if (r != hpk_done) {
 		vtc_log(hp->vl, hp->fatal ? 4 : 0,
-				"Header decoding failed (%d) %d", r, hp->fatal);
-	else if (n == MAX_HDR)
+		    "Header decoding failed (%d) %d", r, hp->fatal);
+	} else if (n == MAX_HDR) {
 		vtc_log(hp->vl, hp->fatal,
-				"Max number of headers reached (%d)", MAX_HDR);
+		    "Max number of headers reached (%d)", MAX_HDR);
+	}
 
 	HPK_FreeIter(iter);
 }
@@ -804,7 +802,7 @@ receive_frame(void *priv)
 			vtc_fatal(s->vl, "Expected CONTINUATION frame for "
 			    "stream %u", expect_cont);
 
-		/* parse the frame according to it type, and fill the metada */
+		/* parse the frame according to it type, and fill the metadata */
 		switch (f->type) {
 			case TYPE_DATA:
 				parse_data(s, f);
@@ -1011,7 +1009,7 @@ cmd_var_resolve(const struct stream *s, const char *spec, char *buf)
 	/* SECTION: stream.spec.zexpect.settings SETTINGS specific
 	 *
 	 * settings.ack
-	 *	"true" if the ACK flag was set, else ""false.
+	 *	"true" if the ACK flag was set, else "false".
 	 *
 	 * settings.push
 	 *	"true" if the push settings was set to yes, "false" if set to
@@ -1260,7 +1258,7 @@ cmd_var_resolve(const struct stream *s, const char *spec, char *buf)
 		else
 			return (NULL);
 	}
-#define H2_ERROR(U,v,sc,r,t) \
+#define H2_ERROR(U,v,sc,g,r,t) \
 	if (!strcmp(spec, #U)) { return (#v); }
 #include "tbl/h2_error.h"
 	return (spec);
@@ -2632,9 +2630,17 @@ stream_thread(void *priv)
 static struct stream *
 stream_new(const char *name, struct http *h)
 {
-	char *p;
+	char *p, buf[20];
 	struct stream *s;
 
+	if (!strcmp("next", name)) {
+		if (h->last_stream > 0)
+			bprintf(buf, "%d", h->last_stream + 2);
+		else
+			bprintf(buf, "%d", 1);
+		name = buf;
+	}
+
 	ALLOC_OBJ(s, STREAM_MAGIC);
 	AN(s);
 	PTOK(pthread_cond_init(&s->cond, NULL));
@@ -2656,6 +2662,7 @@ stream_new(const char *name, struct http *h)
 
 	CHECK_OBJ_NOTNULL(h, HTTP_MAGIC);
 	s->hp = h;
+	h->last_stream = s->id;
 
 	//bprintf(s->connect, "%s", "${v1_sock}");
 	PTOK(pthread_mutex_lock(&h->mtx));
@@ -2761,7 +2768,8 @@ stream_run(struct stream *s)
  *	stream ID [SPEC] [ACTION]
  *
  * ID is the HTTP/2 stream number, while SPEC describes what will be
- * done in that stream.
+ * done in that stream. If ID has the value ``next``, the actual stream
+ * number is computed based on the last one.
  *
  * Note that, when parsing a stream action, if the entity isn't operating
  * in HTTP/2 mode, these spec is ran before::
@@ -2830,7 +2838,7 @@ cmd_stream(CMD_ARGS)
 			continue;
 		}
 		if (**av == '-')
-			vtc_fatal(vl, "Unknown client argument: %s", *av);
+			vtc_fatal(vl, "Unknown stream argument: %s", *av);
 		REPLACE(s->spec, *av);
 	}
 }
